@@ -1,7 +1,7 @@
 #
 # Reportbug module - common functions for reportbug and greportbug
 #   Written by Chris Lawrence <lawrencc@debian.org>
-#   Copyright (C) 1999-2005 Chris Lawrence
+#   Copyright (C) 1999-2006 Chris Lawrence
 #
 # This program is freely distributable per the following license:
 #
@@ -21,17 +21,19 @@
 #
 # Version ##VERSION##; see changelog for revision history
 #
-# $Id: reportbug.py,v 1.28 2005/06/16 21:41:04 lawrencc Exp $
+# $Id: reportbug.py,v 1.35.2.22 2007/04/26 03:48:31 lawrencc Exp $
 
 VERSION = "reportbug ##VERSION##"
 VERSION_NUMBER = "##VERSION##"
-COPYRIGHT = VERSION + '\nCopyright (C) 1999-2004 Chris Lawrence <lawrencc@debian.org>'
+COPYRIGHT = VERSION + '\nCopyright (C) 1999-2006 Chris Lawrence <lawrencc@debian.org>'
 
 import time, sys, os, locale, re, pwd, commands, shlex, debianbts, rfc822
 import socket
 import pprint
+import subprocess
 
 from string import ascii_letters, digits
+from rbtempfile import TempFile, tempfile_prefix
 
 # Paths for dpkg
 DPKGLIB = '/var/lib/dpkg'
@@ -42,11 +44,18 @@ STATUSDB = os.path.join(DPKGLIB, 'status')
 PSEUDOHEADERS = ('Package', 'Version', 'Severity', 'File', 'Tags',
                  'Justification', 'Followup-For', 'Owner')
 
-VALID_UIS = ('newt', 'text', 'gnome')
-#VALID_UIS = ('text', 'gnome')
+VALID_UIS = ['newt', 'text', 'gnome2', 'urwid']
+AVAILABLE_UIS = []
+for ui in VALID_UIS:
+    pkgname = 'reportbug_ui_%s.py' % ui
+    for d in sys.path:
+        if os.path.exists(os.path.join(d, pkgname)):
+            AVAILABLE_UIS.append(ui)
+            break
 
-UIS = {'text': 'A text oriented (console) interface',
-       'gnome': 'A graphical (GNOME) interface'}
+UIS = {'text': 'A text-oriented console interface',
+       'urwid': 'A menu-based console interface',
+       'gnome2': 'A graphical (Gnome 2) interface'}
 
 MODES = {'novice': 'Offer simple prompts, bypassing technical questions.',
          'standard': 'Offer more extensive prompts, including asking about '
@@ -64,30 +73,6 @@ del mode
 
 NEWBIELINE = '*** Please type your report below this line ***'
 
-def rfcdatestr(timeval = None):
-    """Cheesy implementation of an RFC 822 date."""
-    if timeval is None:
-        timeval = time.time()
-
-    try:
-        tm = time.localtime(timeval)
-        if tm[8] > 0:
-            hrs, mins = divmod(-time.altzone/60, 60)
-        else:
-            hrs, mins = divmod(-time.timezone/60, 60)
-
-        # Correct for rounding down
-        if mins and hrs < 0:
-            hrs += 1
-
-        current = locale.setlocale(locale.LC_TIME)
-        locale.setlocale(locale.LC_TIME, "C")
-        ret = time.strftime('%a, %d %b %Y %H:%M:%S', tm)
-        locale.setlocale(locale.LC_TIME, current)
-        return ret + (" %+03d%02d" % (hrs, mins))
-    except:
-        return 'Invalid date: '+`timeval`
-
 fhs_directories = ['/', '/usr', '/usr/share', '/var', '/usr/X11R6',
                    '/usr/man', '/usr/doc', '/usr/bin']
 
@@ -113,16 +98,16 @@ pathdirs = ['/usr/sbin', '/usr/bin', '/sbin', '/bin', '/usr/X11R6/bin',
             '/usr/games']
 
 def search_path_for(filename):
-    dir, file = os.path.split(filename)
-    if dir: return realpath(filename)
+    d, f = os.path.split(filename)
+    if d: return realpath(filename)
     
     path = os.environ.get("PATH", os.defpath).split('/')
-    for dir in pathdirs:
-        if not dir in path:
-            path.append(dir)
+    for d in pathdirs:
+        if not d in path:
+            path.append(d)
     
-    for dir in path:
-        fullname = os.path.join(dir, filename)
+    for d in path:
+        fullname = os.path.join(d, f)
         if os.path.exists(fullname):
             return realpath(fullname)
     return None
@@ -131,21 +116,22 @@ def glob_escape(filename):
     filename = re.sub(r'([*?\[\]])', r'\\\1', filename)
     return filename
 
-def search_pipe(searchfile):
+def search_pipe(searchfile, use_dlocate=True):
     arg = commands.mkarg(searchfile)
-    if os.path.exists('/usr/bin/dlocate'):
+    if use_dlocate and os.path.exists('/usr/bin/dlocate'):
         pipe = os.popen('COLUMNS=79 dlocate -S %s 2>/dev/null' % arg)
     else:
+        use_dlocate = False
         pipe = os.popen('COLUMNS=79 dpkg --search %s 2>/dev/null' % arg)
-    return pipe
+    return (pipe, use_dlocate)
 
-def query_dpkg_for(filename):
+def query_dpkg_for(filename, use_dlocate=True):
     try:
         x = os.getcwd()
     except OSError:
         os.chdir('/')
     searchfilename = glob_escape(filename)
-    pipe = search_pipe(searchfilename)
+    (pipe, dlocate_used) = search_pipe(searchfilename)
     packages = {}
 
     for line in pipe:
@@ -162,6 +148,10 @@ def query_dpkg_for(filename):
             else:
                 packages[package] = [path]
     pipe.close()
+    # Try again without dlocate if no packages found
+    if not packages and dlocate_used:
+        return query_dpkg_for(filename, use_dlocate=False)
+    
     return filename, packages
 
 def find_package_for(filename, pathonly=False):
@@ -202,7 +192,7 @@ def get_email_addr(addr):
 def get_email(email='', realname=''):
     return get_email_addr(get_user_id(email, realname))
 
-def get_user_id(email='', realname=''):
+def get_user_id(email='', realname='', charset='utf-8'):
     uid = os.getuid()
     info = pwd.getpwuid(uid)
     email = (os.environ.get('REPORTBUGEMAIL', email) or
@@ -232,11 +222,19 @@ def get_user_id(email='', realname=''):
 
     if not realname:
         return email
+
+    # Decode the realname from the charset -
+    # but only if it is not already in Unicode
+    if isinstance(realname, str):
+        realname = realname.decode(charset, 'replace')
     
     if re.match(r'[\w\s]+$', realname):
-        return '%s <%s>' % (realname, email)
+        return u'%s <%s>' % (realname, email)
 
-    return rfc822.dump_address_pair( (realname, email) )
+    addr = rfc822.dump_address_pair( (realname, email) )
+    if isinstance(addr, str):
+        addr = addr.decode('utf-8', 'replace')
+    return addr
 
 statuscache = {}
 def get_package_status(package, avail=False):
@@ -279,6 +277,9 @@ def get_package_status(package, avail=False):
         output = commands.getoutput(
             "COLUMNS=79 dpkg --status %s 2>/dev/null" % packarg)
 
+    # dpkg output is in UTF-8 format
+    output = output.decode('utf-8', 'replace')
+
     for line in output.split(os.linesep):
         line = line.rstrip()
         if not line: continue
@@ -352,67 +353,114 @@ def get_package_status(package, avail=False):
         statuscache[package] = info
     return info
 
-dbase = []
-avail = []
+#dbase = []
+#avail = []
+
+# Object that essentially chunkifies the output of apt-cache dumpavail
+class AvailDB(object):
+    def __init__(self, fp=None, popenob=None):
+        self.popenob = popenob
+        if fp:
+            self.fp = fp
+        elif popenob:
+            self.fp = popenob.stdout
+
+    def __iter__(self):
+        return self
+
+    def next(self):
+        chunk = u''
+        while True:
+            if self.popenob:
+                if self.popenob.returncode:
+                    break
+
+            line = self.fp.readline()
+            if not line:
+                break
+                
+            if line == '\n':
+                return chunk
+            chunk += line.decode('utf-8', 'replace')
+
+        if chunk:
+            return chunk
+
+        raise StopIteration
+
+    def __del__(self):
+        #print >> sys.stderr, 'availdb cleanup', repr(self.popenob), repr(self.fp)
+        if self.popenob:
+            # Clear the pipe before shutting it down
+            while True:
+                if self.popenob.returncode:
+                    break
+                stuff = self.fp.read(65536)
+                if not stuff:
+                    break
+            self.popenob.wait()
+        if self.fp:
+            self.fp.close()
+
 def get_dpkg_database():
-    global dbase
+    if os.path.exists(STATUSDB):
+        fp = open(STATUSDB)
+        if fp:
+            return AvailDB(fp=fp)
 
-    if not dbase:
-        if os.path.exists(STATUSDB):
-            dbase = file(STATUSDB).read().split('\n\n')
-
-    if not dbase:
-        print >> sys.stderr, 'Unable to open', STATUSDB
-        sys.exit(1)
-
-    return dbase
+    print >> sys.stderr, 'Unable to open', STATUSDB
+    sys.exit(1)
 
 def get_avail_database():
-    global avail
+    #print >> sys.stderr, 'Searching available database'
+    subp = subprocess.Popen(('apt-cache', 'dumpavail'), stdout=subprocess.PIPE)
+    return AvailDB(popenob=subp)
 
-    if not avail:
-        avail = commands.getoutput('apt-cache dumpavail 2>/dev/null').split('\n\n')
-        if not avail:
-            if os.path.exists(AVAILDB):
-                avail = file(AVAILDB).read().split('\n\n')
+def available_package_description(package):
+    data = commands.getoutput('apt-cache show'+commands.mkarg(package))
+    data = data.decode('utf-8', 'replace')
+    descre = re.compile(r'^Description: (.*)$')
+    for line in data.split('\n'):
+        m = descre.match(line)
+        if m:
+            return m.group(1)
+    return None
 
-    if not avail:
-        print >> sys.stderr, 'Unable to open', AVAILDB
-        sys.exit(1)
-
-    return avail
-
-def get_source_package(package):
-    """Return any binary packages provided by a source package."""
-    packinfo = get_avail_database()
+def get_source_name(package):
     packages = []
-    packob = re.compile(r'^Package: (?P<pkg>.*)$', re.MULTILINE)
-    descob = re.compile(r'^Description: (?P<desc>.*)$', re.MULTILINE)
-    if package == 'libc':
-        searchob = re.compile(r'^(Package|Source): g?libc[\d.]*$',
-                              re.MULTILINE)
-    else:
-        searchob = re.compile(r'^(Package|Source): '+re.escape(package)+r'$',
-                              re.MULTILINE)
+
+    data = commands.getoutput('apt-cache showsrc'+commands.mkarg(package))
+    data = data.decode('utf-8', 'replace')
+    packre = re.compile(r'^Package: (.*)$')
+    for line in data.split('\n'):
+        m = packre.match(line)
+        if m:
+            return m.group(1)
+    return None
+        
+def get_source_package(package):
+    packages = []
+    retlist = []
+
+    data = commands.getoutput('apt-cache showsrc'+commands.mkarg(package))
+    data = data.decode('utf-8', 'replace')
+    binre = re.compile(r'^Binary: (.*)$')
+    for line in data.split('\n'):
+        m = binre.match(line)
+        if m:
+            packs = m.group(1)
+            packlist = re.split(r',\s*', packs)
+            packages += packlist
     
-    for p in packinfo:
-        match = searchob.search(p)
-        if match:
-            packname = packdesc = ''
-            namematch, descmatch = packob.search(p), descob.search(p)
+    for p in packages:
+        desc = available_package_description(p)
+        if desc:
+            retlist += [(p, desc)]
 
-            if namematch:
-                packname = namematch.group('pkg')
-            if descmatch:
-                packdesc = descmatch.group('desc')
-            
-            if packname:
-                packages.append( (packname, packdesc) )
+    retlist.sort()
+    return retlist
 
-    packages.sort()
-    return packages
-
-def get_package_info(packages):
+def get_package_info(packages, skip_notfound=False):
     if not packages:
         return []
     
@@ -473,6 +521,9 @@ def get_package_info(packages):
                 if provides not in found:
                     found[provides] = True
 
+    if skip_notfound:
+        return ret
+
     for group in groups:
         notfound = [x for x in group if x not in found]
         if len(notfound) == len(group):
@@ -483,12 +534,11 @@ def get_package_info(packages):
     return ret
 
 def packages_providing(package):
-    aret = get_package_info([((package,), package)])
+    aret = get_package_info([((package,), package)], skip_notfound=True)
     ret = []
     for pkg in aret:
         ret.append( (pkg[0], pkg[3]) )
-        if not pkg[2]: return []
-    
+
     return ret
 
 def get_dependency_info(package, depends, rel="depends on"):
@@ -580,7 +630,9 @@ def get_debian_release_info():
             debvers = dists[0][1]
 
     if os.path.exists('/etc/debian_version'):
-        verfile = file('/etc/debian_version').readline().strip()
+        fob = open('/etc/debian_version')
+        verfile = fob.readline().strip()
+        fob.close()
 
     if verfile:
         debinfo += 'Debian Release: '+verfile+'\n'
@@ -613,23 +665,9 @@ def generate_blank_report(package, pkgversion, severity, justification,
                           pseudos=None):
     un = os.uname()
     utsmachine = un[4]
-    debinfo = ''
+    debinfo = u''
     shellpath = realpath('/bin/sh')
 
-    if debianbts.SYSTEMS[system].get('query-dpkg', True):
-        debinfo += get_debian_release_info()
-        debarch = get_arch()
-        if debarch:
-            if utsmachine == debarch:
-                debinfo += 'Architecture: %s\n' % (debarch)
-            else:
-                debinfo += 'Architecture: %s (%s)\n' % (debarch, utsmachine)
-        else:
-            debinfo += 'Architecture: ? (%s)\n' % utsmachine
-
-    if shellpath != '/bin/sh':
-        debinfo += 'Shell:  /bin/sh linked to %s\n' % shellpath
-
     locinfo = []
     langsetting = os.environ.get('LANG', 'C')
     allsetting = os.environ.get('LC_ALL', '')
@@ -651,28 +689,28 @@ def generate_blank_report(package, pkgversion, severity, justification,
         package = debianbts.SYSTEMS[system]['namefmt'] % package
 
     if pseudos:
-        headers = '\n'.join(pseudos)+'\n'
+        headers = u'\n'.join(pseudos)+u'\n'
     else:
-        headers = ''
+        headers = u''
     
     if pkgversion:
-        headers += 'Version: %s\n' % pkgversion
+        headers += u'Version: %s\n' % pkgversion
 
     if not exinfo:
         if severity:
-            headers += 'Severity: %s\n' % severity
+            headers += u'Severity: %s\n' % severity
 
         if justification:
-            headers += 'Justification: %s\n' % justification
+            headers += u'Justification: %s\n' % justification
 
         if tags:
-            headers += 'Tags: %s\n' % tags
+            headers += u'Tags: %s\n' % tags
 
         if foundfile:
-            headers += 'File: %s\n' % foundfile
+            headers += u'File: %s\n' % foundfile
 
     if mode < MODE_ADVANCED:
-        body = NEWBIELINE+'\n\n'+body
+        body = NEWBIELINE+u'\n\n'+body
     
     report = "\n"
     if not exinfo:
@@ -704,16 +742,66 @@ def generate_blank_report(package, pkgversion, severity, justification,
             exinfo, package, headers)
 
     if not body:
-        body = "\n"
+        body = u"\n"
 
-    uname_string = '%s %s' % (un[0], un[2])
+    if debianbts.SYSTEMS[system].get('query-dpkg', True):
+        debinfo += get_debian_release_info()
+        debarch = get_arch()
+        if debarch:
+            if utsmachine == debarch:
+                debinfo += u'Architecture: %s\n' % (debarch)
+            else:
+                debinfo += u'Architecture: %s (%s)\n' % (debarch, utsmachine)
+        else:
+            debinfo += u'Architecture: ? (%s)\n' % utsmachine
+        debinfo += u'\n'
 
-    return """%s%s%s
+    if un[0] == 'GNU':
+        # Use uname -v on Hurd
+        uname_string = un[3]
+    else:
+        kern = un[0]
+        if kern.startswith('GNU/'):
+            kern = kern[4:]
+
+        uname_string = '%s %s' % (kern, un[2])
+        if kern == 'Linux':
+            kinfo = []
+
+            if 'SMP' in un[3]:
+                cores = get_cpu_cores()
+                if cores > 1:
+                    kinfo += ['SMP w/%d CPU cores' % cores]
+                else:
+                    kinfo += ['SMP w/1 CPU core']
+            if 'PREEMPT' in un[3]:
+                kinfo += ['PREEMPT']
+
+            if kinfo:
+                uname_string = '%s (%s)' % (uname_string, '; '.join(kinfo))
+
+    if uname_string:
+        debinfo += u'Kernel: %s\n' % uname_string
+
+    if locinfo:
+        debinfo += u'Locale: %s\n' % locinfo
+    if shellpath != '/bin/sh':
+        debinfo += u'Shell: /bin/sh linked to %s\n' % shellpath
+
+    return u"""%s%s%s
 -- System Information:
-%sKernel: %s
-Locale: %s
-%s%s""" % (report, body, incfiles, debinfo, uname_string, locinfo,
-           depinfo, confinfo)
+%s%s%s""" % (report, body, incfiles, debinfo, depinfo, confinfo)
+
+def get_cpu_cores():
+    cpucount = 0
+    fob = open('/proc/cpuinfo')
+    for line in fob:
+        if line.startswith('processor'):
+            cpucount += 1
+            #print repr(line), cpucount
+    fob.close()
+
+    return max(cpucount, 1)
 
 class our_lex(shlex.shlex):
     def get_token(self):
@@ -782,7 +870,8 @@ def parse_config_files():
                 elif token in ('email', 'realname', 'replyto', 'http_proxy',
                                'smtphost', 'editor', 'mua', 'mta', 'smtpuser',
                                'smtppasswd', 'justification', 'keyid'):
-                    args[token] = lex.get_token()
+                    bit = lex.get_token()
+                    args[token] = bit.decode('utf-8', 'replace')
                 elif token in ('no-smtptls', 'smtptls'):
                     args['smtptls'] = (token == 'smtptls')
                 elif token == 'sign':
@@ -795,7 +884,7 @@ def parse_config_files():
                         args['sign'] = ''
                 elif token == 'ui':
                     token = lex.get_token().lower()
-                    if token in VALID_UIS:
+                    if token in AVAILABLE_UIS:
                         args['interface'] = token
                 elif token == 'mode':
                     arg = lex.get_token().lower()
